@@ -8,13 +8,17 @@
 
 import sys
 import os
-from typing import Iterable, List, Any
+from typing import Iterable, List, Any, Union
 from abc import abstractmethod, ABC
 from enum import Enum
+import io
 
 from .paths import Path
 from .size import Size
 from .serialize import load, dump
+from .serialize_ctx import load_context, Context
+from .tree import ContentTree
+
 
 class NodeType(Enum):
     file = 1
@@ -117,6 +121,10 @@ class NodeInfo(ABC):
         '''
         return False
 
+    def is_symlink(self):
+        ''' get whether the node is a symbolic link node. '''
+        return os.path.islink(self._path)
+
     # abstract methods
 
     @abstractmethod
@@ -129,6 +137,11 @@ class NodeInfo(ABC):
         ''' create hardlink for the node. '''
         raise NotImplementedError
 
+    @abstractmethod
+    def create_symlink(self, dest_path: str):
+        ''' create symbolic link for the node. '''
+        raise NotImplementedError
+
 
 class FileInfo(NodeInfo):
 
@@ -141,6 +154,14 @@ class FileInfo(NodeInfo):
                     newline=newline,
                     closefd=closefd)
 
+    def open_for_read_bytes(self, *, buffering=-1):
+        ''' open the file with read bytes mode. '''
+        return self.open('rb', buffering=buffering)
+
+    def open_for_read_text(self, *, encoding='utf-8'):
+        ''' open the file with read text mode. '''
+        return self.open('r', encoding=encoding)
+
     @property
     def size(self):
         ''' get file size. '''
@@ -149,12 +170,30 @@ class FileInfo(NodeInfo):
     def write(self, data, *, mode=None, buffering=-1, encoding=None, newline=None):
         ''' write data into the file. '''
         if mode is None:
-            mode = 'w' if isinstance(data, str) else 'wb'
+            if isinstance(data, (str, io.TextIOBase)):
+                mode = 'w'
+            elif isinstance(data, (bytes, bytearray, io.BufferedIOBase)):
+                mode = 'wb'
+            else:
+                raise TypeError(type(data))
+
         with self.open(mode=mode, buffering=buffering, encoding=encoding, newline=newline) as fp:
-            return fp.write(data)
+            if isinstance(data, (str, bytes, bytearray)):
+                return fp.write(data)
+            else:
+                read_buffering = buffering
+                if read_buffering < 2:
+                    read_buffering = io.DEFAULT_BUFFER_SIZE
+                total = 0
+                while True:
+                    b = data.read(read_buffering)
+                    if not b:
+                        break
+                    total += fp.write(b)
+                return total
 
     def read(self, mode='r', *, buffering=-1, encoding=None, newline=None):
-        ''' read data from the file. '''
+        ''' read all content from the file. '''
         with self.open(mode=mode, buffering=buffering, encoding=encoding, newline=newline) as fp:
             return fp.read()
 
@@ -163,11 +202,21 @@ class FileInfo(NodeInfo):
         mode = 'a' if append else 'w'
         return self.write(text, mode=mode, encoding=encoding)
 
-    def write_bytes(self, data: bytes, *, append=True):
+    def write_bytes(self, data: Union[bytes, bytearray], *, append=True):
         ''' write bytes into the file. '''
         mode = 'ab' if append else 'wb'
         return self.write(data, mode=mode)
 
+    def write_from_stream(self, stream: io.IOBase, *, append=True):
+        if not isinstance(stream, io.IOBase):
+            raise TypeError(type(stream))
+        if not stream.readable():
+            raise ValueError('stream is unable to read.')
+        mode = 'a' if append else 'w'
+        if not isinstance(stream, io.TextIOBase):
+            mode += 'b'
+        return self.write(stream, mode=mode)
+
     def copy_to(self, dest, buffering: int = -1):
         '''
         copy the file to dest path.
@@ -193,14 +242,51 @@ class FileInfo(NodeInfo):
 
     def read_text(self, encoding='utf-8') -> str:
         ''' read all text into memory. '''
-        with self.open('r', encoding=encoding) as fp:
+        with self.open_for_read_text(encoding=encoding) as fp:
             return fp.read()
 
     def read_bytes(self) -> bytes:
         ''' read all bytes into memory. '''
-        with self.open('rb') as fp:
+        with self.open_for_read_bytes() as fp:
             return fp.read()
 
+    def read_into_stream(self, stream: io.IOBase, *, encoding=None, buffering: int = -1):
+        ''' read all content into stream. '''
+        if not isinstance(stream, io.IOBase):
+            raise TypeError(type(stream))
+        if not stream.writable():
+            raise ValueError('stream is unable to write.')
+
+        if buffering < 0:
+            buffering = io.DEFAULT_BUFFER_SIZE
+
+        if isinstance(stream, io.TextIOBase):
+            encoding = encoding or 'utf-8'
+            fp = self.open_for_read_text(encoding=encoding)
+        else:
+            fp = self.open_for_read_bytes()
+
+        with fp:
+            while True:
+                data = fp.read(buffering)
+                if not data:
+                    break
+                stream.write(data)
+
+    def __iadd__(self, other: Union[str, bytes, bytearray, io.IOBase, 'FileInfo']):
+        if isinstance(other, str):
+            self.write_text(other)
+        elif isinstance(other, (bytes, bytearray)):
+            self.write_bytes(other)
+        elif isinstance(other, io.IOBase):
+            self.write_from_stream(other)
+        elif isinstance(other, FileInfo):
+            with other.open_for_read_bytes() as fp:
+                self.write_from_stream(fp)
+        else:
+            raise TypeError(type(other))
+        return self
+
     # override common methods
 
     @property
@@ -224,6 +310,10 @@ class FileInfo(NodeInfo):
         ''' create hardlink for the file. '''
         os.link(self._path, dest_path)
 
+    def create_symlink(self, dest_path: str):
+        ''' create symbolic link for the file. '''
+        os.symlink(self._path, dest_path, target_is_directory=False)
+
     # load/dump system.
 
     def load(self, format=None, *, kwargs={}):
@@ -247,25 +337,24 @@ class FileInfo(NodeInfo):
         '''
         return dump(self, obj, format=format, kwargs=kwargs)
 
-    def load_session(self, format=None, *, load_kwargs={}, dump_kwargs={}):
+    def load_context(self, format=None, *, load_kwargs={}, dump_kwargs={}, lock=False) -> Context:
         '''
-        load the file in a `session`, auto dump `session.data` into file when session `__exit__`;
+        load the file in a context, auto dump `context.data` into file when context exit.
 
-        - if the file does not exists, `session.data` will be `None`.
-        - set `session.data` to `None` will remove the file from disk.
+        - if the file does not exists, `context.data` will be `None`.
+        - set `context.data` to `None` will remove the file from disk.
+        - by default, `context.save_on_exit` is `True`.
+        - if `lock` is `True`, lock the file until the context exit.
 
         usage:
 
         ``` py
-        with file.load_session() as session:
-            # read or edit session.data
+        with file.load_context() as context:
+            # read or edit context.data
             ...
         ```
         '''
-        data = self.load(format, kwargs=load_kwargs) if self.is_file() else None
-        session = _LoadSession(self, format, dump_kwargs)
-        session.data = data
-        return session
+        return load_context(self, format, load_kwargs=load_kwargs, dump_kwargs=dump_kwargs, lock=lock)
 
     # hash system
 
@@ -277,8 +366,14 @@ class FileInfo(NodeInfo):
 
         for example: `get_file_hash('md5', 'sha1')` return `('XXXX1', 'XXXX2')`
         '''
-        from .hashs import hashfile_hexdigest
-        return hashfile_hexdigest(self._path, algorithms)
+        with self.get_hasher(*algorithms) as hasher:
+            while hasher.read_block():
+                pass
+            return hasher.result
+
+    def get_hasher(self, *algorithms: str):
+        from .hashs import Hasher
+        return Hasher(self._path, algorithms)
 
 
 class DirectoryInfo(NodeInfo):
@@ -358,15 +453,24 @@ class DirectoryInfo(NodeInfo):
 
     # tree api
 
-    def get_tree(self) -> dict:
+    def get_tree(self, *, as_stream=False) -> ContentTree:
         '''
         Get structure tree from current directory.
+
+        if `as_stream` is `True`,
+        collect all files as `file-object` instead of read entire file into memory,
+        you need to call `__exit__` after you used it.
         '''
-        tree = {}
+        tree = ContentTree()
         for item in self.list_items():
             name = str(item.path.name)
             if item.node_type == NodeType.file:
-                tree[name] = item.read(mode='rb')
+                if as_stream:
+                    fp = item.open_for_read_bytes()
+                    tree[name] = fp
+                    tree.enter(fp)
+                else:
+                    tree[name] = item.read(mode='rb')
             else:
                 tree[name] = item.get_tree()
         return tree
@@ -393,7 +497,7 @@ class DirectoryInfo(NodeInfo):
             if not isinstance(key, str):
                 raise TypeError(key)
 
-            if isinstance(value, (str, bytes)):
+            if isinstance(value, (str, bytes, bytearray, io.IOBase)):
                 subfile = self.get_fileinfo(key)
                 if subfile.is_file():
                     if mode == 0:
@@ -402,10 +506,15 @@ class DirectoryInfo(NodeInfo):
                         raise FileExistsError(subfile.path)
                     elif mode == 2:
                         subfile.delete()
+
                 if isinstance(value, str):
-                    subfile.write_text(value)
+                    subfile.write_text(value, append=False)
+
+                elif isinstance(value, (bytes, bytearray)):
+                    subfile.write_bytes(value, append=False)
+
                 else:
-                    subfile.write_bytes(value)
+                    subfile.write_from_stream(value, append=False)
 
             elif isinstance(value, dict):
                 subdir = self.get_dirinfo(key)
@@ -445,26 +554,6 @@ class DirectoryInfo(NodeInfo):
         for item in self.list_items():
             item.create_hardlink(os.path.join(dest_path, item.path.name))
 
-
-class _LoadSession:
-    '''
-    a loaded session for a `FileInfo`.
-
-    note: set `.data` to `None` will remove the file from disk.
-    '''
-
-    def __init__(self, file: FileInfo, format, dump_kwargs: dict):
-        self.data: Any = None
-        self._file = file
-        self._format = format
-        self._dump_kwargs = dump_kwargs
-
-    def __enter__(self):
-        return self
-
-    def __exit__(self, *_):
-        if self.data is None:
-            if self._file.is_file():
-                self._file.delete()
-        else:
-            self._file.dump(self.data, self._format, kwargs=self._dump_kwargs)
+    def create_symlink(self, dest_path: str):
+        ''' create symbolic link for the directory. '''
+        os.symlink(self._path, dest_path, target_is_directory=True)
